/* (C) 2009-2010 Yana */

#if !defined(YK0A_H)
#define YK0A_H

#include "yk0a_base.h"

#define YK0A_CONTEXT_R0 0
#define YK0A_CONTEXT_R1 1
#define YK0A_CONTEXT_R2 2
#define YK0A_CONTEXT_R3 3
#define YK0A_CONTEXT_R4 4
#define YK0A_CONTEXT_R5 5
#define YK0A_CONTEXT_R6 6
#define YK0A_CONTEXT_R7 7
#define YK0A_CONTEXT_R8 8
#define YK0A_CONTEXT_R9 9
#define YK0A_CONTEXT_R10 10
#define YK0A_CONTEXT_R11 11
#define YK0A_CONTEXT_R12 12
#define YK0A_CONTEXT_R13 13 /* sp */
#define YK0A_CONTEXT_R14 14 /* lr */
#define YK0A_CONTEXT_R15 15 /* pc */
#define YK0A_CONTEXT_SPSR 16
#define YK0A_CONTEXT_CPSR 17

#define YK0A_FLAG_KERNEL_MASK (3 << YK0A_FLAG_KERNEL_SHIFT)
#define YK0A_FLAG_KERNEL_SHIFT 30

#define YK0A_FLAG_KERNEL_RECEIVER (2 << YK0A_FLAG_KERNEL_SHIFT)
#define YK0A_FLAG_KERNEL_SENDER (1 << YK0A_FLAG_KERNEL_SHIFT)

#define YK0A_PROCEDURE_ACCEPT 0
#define YK0A_PROCEDURE_BIT_FLAG 1
#define YK0A_PROCEDURE_CALL 2
#define YK0A_PROCEDURE_CONNECT 3
#define YK0A_PROCEDURE_FORWARD 4
#define YK0A_PROCEDURE_GET_CONTEXT 5
#define YK0A_PROCEDURE_GET_PRIORITY 6
#define YK0A_PROCEDURE_LISTEN 7
#define YK0A_PROCEDURE_REPLY 8
#define YK0A_PROCEDURE_SET_CONTEXT 9
#define YK0A_PROCEDURE_SET_PRIORITY 10

#if !defined(YK0A_PREPROCESS_ONLY)

#include <stddef.h>

typedef yk0a_uint32_t yk0a_context_type_t;
typedef yk0a_uint32_t yk0a_context_value_t;
typedef yk0a_uint32_t yk0a_error_t;
typedef yk0a_uint32_t yk0a_flag_t;
typedef yk0a_uint32_t yk0a_port_t;
typedef yk0a_uint32_t yk0a_priority_t;
typedef yk0a_uint32_t yk0a_task_t;

__inline__ static yk0a_error_t yk0a_accept(yk0a_port_t port, yk0a_task_t *task) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_ACCEPT;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1) : "memory");
	if (task != NULL) {
		*task = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_bit_flag(yk0a_port_t port, yk0a_flag_t flag_off, yk0a_flag_t flag_on, yk0a_flag_t *flag_out) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_BIT_FLAG;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = flag_off;
	register yk0a_uint32_t r3 __asm__("r3") = flag_on;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2), "r"(r3) : "memory");
	if (flag_out != NULL) {
		*flag_out = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_call(yk0a_port_t port) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1) : "memory");
	return r0;
}

__inline__ static yk0a_error_t yk0a_connect(yk0a_task_t task_replyee, yk0a_task_t task_replyer) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CONNECT;
	register yk0a_uint32_t r1 __asm__("r1") = task_replyee;
	register yk0a_uint32_t r2 __asm__("r2") = task_replyer;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2) : "memory");
	return r0;
}

__inline__ static yk0a_error_t yk0a_forward(yk0a_port_t port, yk0a_task_t task) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_FORWARD;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = task;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2) : "memory");
	return r0;
}

__inline__ static yk0a_error_t yk0a_get_context(yk0a_task_t task, yk0a_context_type_t type, yk0a_context_value_t *value) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_GET_CONTEXT;
	register yk0a_uint32_t r1 __asm__("r1") = task;
	register yk0a_uint32_t r2 __asm__("r2") = type;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2) : "memory");
	if (value != NULL) {
		*value = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_get_priority(yk0a_task_t task, yk0a_priority_t *priority) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_GET_PRIORITY;
	register yk0a_uint32_t r1 __asm__("r1") = task;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1) : "memory");
	if (priority != NULL) {
		*priority = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_listen(yk0a_port_t port, yk0a_flag_t flag_wait, yk0a_flag_t *flag_out) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_LISTEN;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = flag_wait;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2) : "memory");
	if (flag_out != NULL) {
		*flag_out = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_reply(yk0a_task_t task) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_REPLY;
	register yk0a_uint32_t r1 __asm__("r1") = task;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1) : "memory");
	return r0;
}

__inline__ static yk0a_error_t yk0a_set_context(yk0a_task_t task, yk0a_context_type_t type, yk0a_context_value_t value_in, yk0a_context_value_t *value_out) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_SET_CONTEXT;
	register yk0a_uint32_t r1 __asm__("r1") = task;
	register yk0a_uint32_t r2 __asm__("r2") = type;
	register yk0a_uint32_t r3 __asm__("r3") = value_in;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2), "r"(r3) : "memory");
	if (value_out != NULL) {
		*value_out = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_set_priority(yk0a_task_t task, yk0a_priority_t priority_in, yk0a_priority_t *priority_out) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_SET_PRIORITY;
	register yk0a_uint32_t r1 __asm__("r1") = task;
	register yk0a_uint32_t r2 __asm__("r2") = priority_in;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2) : "memory");
	if (priority_out != NULL) {
		*priority_out = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t yk0a_get_flag(yk0a_port_t port, yk0a_flag_t *flag) {
	return yk0a_bit_flag(port, 0, 0, flag);
}

__inline__ static yk0a_error_t yk0a_set_flag(yk0a_port_t port, yk0a_flag_t flag_in, yk0a_flag_t *flag_out) {
	return yk0a_bit_flag(port, ~0, flag_in, flag_out);
}

#endif

#endif
