/* (C) 2009-2010 Yana */

#include "device.h"
#include "start_setup.h"
#include "tool.h"
#include "yk0a.h"
#include "yk0a_interrupt.h"
#include "yk0a_setup.h"
#include <stddef.h>

typedef struct {
	tool_queue_t queue;
} wrap_task_control_t;

#define WRAP0_CONS_GETCHAR 0
#define WRAP0_CONS_HALT 1
#define WRAP0_CONS_PUTCHAR 2

__inline__ static yk0a_error_t wrap0_cons_getchar(yk0a_port_t port, yk0a_uint8_t *c) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CONS_GETCHAR;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2) : "memory");
	if (c != NULL) {
		*c = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t wrap0_cons_halt(yk0a_port_t port) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CONS_HALT;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2) : "memory");
	return r0;
}

__inline__ static yk0a_error_t wrap0_cons_putchar(yk0a_port_t port, yk0a_uint8_t c) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CONS_PUTCHAR;
	register yk0a_uint32_t r3 __asm__("r3") = c;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2), "r"(r3) : "memory");
	return r0;
}

#define CONTEXT_R0 YK0A_CONTEXT_R0
#define CONTEXT_R1 YK0A_CONTEXT_R1
#define CONTEXT_R2 YK0A_CONTEXT_R2
#define CONTEXT_R3 YK0A_CONTEXT_R3

wrap_task_control_t wrap_task_control[YK0A_SETUP_TASK_SIZE];

yk0a_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));
yk0a_uint8_t task1_stack[TASK1_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));
yk0a_uint8_t task2_stack[TASK2_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));

void task1(void);
void task2(void);

void task0(void) {
	yk0a_flag_t f;
	yk0a_task_t t;
	tool_task_start(1, task1, task1_stack + TASK1_STACK_SIZE / YK0A_STACK_ALIGN * YK0A_STACK_ALIGN);
	yk0a_set_priority(2, 1, NULL);
	tool_task_start(2, task2, task2_stack + TASK2_STACK_SIZE / YK0A_STACK_ALIGN * YK0A_STACK_ALIGN);
	yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
	tool_cons_put_string("{");
	tool_cons_put_hexadecimal(f);
	if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
		yk0a_uint32_t cpsr;
		yk0a_get_context(t, YK0A_CONTEXT_CPSR, &cpsr);
		tool_cons_put_string(" ");
		tool_cons_put_hexadecimal(t);
		tool_cons_put_string(" ");
		tool_cons_put_hexadecimal(cpsr);
	}
	tool_cons_put_string("}\n");
	device_cons_halt();
}

void task1(void) {
	yk0a_uint8_t c;
	tool_queue_t q;
	yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
	c = 0;
	tool_queue_initialize(&q);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_INTERRUPT_PORT, YK0A_FLAG_KERNEL_SENDER | DEVICE_INTERRUPT_CONS, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_INTERRUPT_PORT, &t) == 0) {
			yk0a_uint32_t r2;
			yk0a_uint32_t r3;
			yk0a_get_context(t, CONTEXT_R2, &r2);
			switch (r2) {
			case WRAP0_CONS_GETCHAR:
				tool_queue_append(&q, (tool_queue_t *)&wrap_task_control[t]);
				break;
			case WRAP0_CONS_PUTCHAR:
				yk0a_get_context(t, CONTEXT_R3, &r3);
				yk0a_reply(t);
				device_cons_putchar(r3);
				break;
			default:
				yk0a_forward(YK0A_SETUP_EXCEPTION_PORT, t);
				break;
			}
		}
		if ((f & DEVICE_INTERRUPT_CONS) != 0) {
			c = device_cons_getchar();
			yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
		}
		if (c != 0 && tool_queue_copy(&q) != NULL) {
			t = (wrap_task_control_t *)tool_queue_remove(&q) - &wrap_task_control[0];
			yk0a_set_context(t, CONTEXT_R1, c, NULL);
			yk0a_reply(t);
			c = 0;
		}
	}
}

void task2(void) {
	while (1) {
		yk0a_uint8_t c;
		wrap0_cons_getchar(YK0A_SETUP_INTERRUPT_PORT, &c);
		if (c < ' ') {
			break;
		}
		wrap0_cons_putchar(YK0A_SETUP_INTERRUPT_PORT, c);
	}
	wrap0_cons_halt(YK0A_SETUP_INTERRUPT_PORT);
}
