/* (C) 2009-2010 Yana */

#include "device.h"
#include "start_setup.h"
#include "tool.h"
#include "yk0a.h"
#include "yk0a_interrupt.h"
#include "yk0a_setup.h"
#include <stddef.h>

#define WRAP_PORT_SIZE 1

typedef yk0a_uint32_t wrap_port_t;

typedef struct {
	tool_queue_t queue;
	yk0a_flag_t flag;
	yk0a_task_t receiver;
} wrap_port_control_t;

typedef struct {
	tool_queue_t queue;
} wrap_task_control_t;

#define WRAP0_ACCEPT 0
#define WRAP0_CALL 1
#define WRAP0_LISTEN 2

__inline__ static yk0a_error_t wrap0_accept(yk0a_port_t port, wrap_port_t wrap_port, yk0a_task_t *task) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_ACCEPT;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2), "r"(r3) : "memory");
	if (task != NULL) {
		*task = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t wrap0_call(yk0a_port_t port, wrap_port_t wrap_port) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CALL;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2), "r"(r3) : "memory");
	return r0;
}

__inline__ static yk0a_error_t wrap0_listen(yk0a_port_t port, wrap_port_t wrap_port, yk0a_flag_t flag_wait, yk0a_flag_t *flag_out) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_LISTEN;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	register yk0a_uint32_t r4 __asm__("r4") = flag_wait;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2), "r"(r3), "r"(r4) : "memory");
	if (flag_out != NULL) {
		*flag_out = r1;
	}
	return r0;
}

#define WRAP1_CONS_GETCHAR 0
#define WRAP1_CONS_HALT 1
#define WRAP1_CONS_PUTCHAR 2

__inline__ static yk0a_error_t wrap1_cons_getchar(yk0a_port_t port, wrap_port_t wrap_port, yk0a_uint8_t *c) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CALL;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	register yk0a_uint32_t r4 __asm__("r4") = WRAP1_CONS_GETCHAR;
	__asm__ __volatile__("swi 0" : "=r"(r0), "=r"(r1) : "0"(r0), "1"(r1), "r"(r2), "r"(r3), "r"(r4) : "memory");
	if (c != NULL) {
		*c = r1;
	}
	return r0;
}

__inline__ static yk0a_error_t wrap1_cons_halt(yk0a_port_t port, wrap_port_t wrap_port) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CALL;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	register yk0a_uint32_t r4 __asm__("r4") = WRAP1_CONS_HALT;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2), "r"(r3), "r"(r4) : "memory");
	return r0;
}

__inline__ static yk0a_error_t wrap1_cons_putchar(yk0a_port_t port, wrap_port_t wrap_port, yk0a_uint8_t c) {
	register yk0a_uint32_t r0 __asm__("r0") = YK0A_PROCEDURE_CALL;
	register yk0a_uint32_t r1 __asm__("r1") = port;
	register yk0a_uint32_t r2 __asm__("r2") = WRAP0_CALL;
	register yk0a_uint32_t r3 __asm__("r3") = wrap_port;
	register yk0a_uint32_t r4 __asm__("r4") = WRAP1_CONS_PUTCHAR;
	register yk0a_uint32_t r5 __asm__("r5") = c;
	__asm__ __volatile__("swi 0" : "=r"(r0) : "0"(r0), "r"(r1), "r"(r2), "r"(r3), "r"(r4), "r"(r5) : "memory");
	return r0;
}

#define CONTEXT_R0 YK0A_CONTEXT_R0
#define CONTEXT_R1 YK0A_CONTEXT_R1
#define CONTEXT_R2 YK0A_CONTEXT_R2
#define CONTEXT_R3 YK0A_CONTEXT_R3
#define CONTEXT_R4 YK0A_CONTEXT_R4
#define CONTEXT_R5 YK0A_CONTEXT_R5

wrap_port_control_t wrap_port_control[WRAP_PORT_SIZE];
wrap_task_control_t wrap_task_control[YK0A_SETUP_TASK_SIZE];

yk0a_uint8_t task0_stack[TASK0_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));
yk0a_uint8_t task1_stack[TASK1_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));
yk0a_uint8_t task2_stack[TASK2_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));
yk0a_uint8_t task3_stack[TASK3_STACK_SIZE] __attribute__((aligned(YK0A_STACK_ALIGN)));

void task1(void);
void task2(void);
void task3(void);

void task0(void) {
	yk0a_flag_t f;
	yk0a_task_t t;
	tool_task_start(1, task1, task1_stack + TASK1_STACK_SIZE / YK0A_STACK_ALIGN * YK0A_STACK_ALIGN);
	tool_task_start(2, task2, task2_stack + TASK2_STACK_SIZE / YK0A_STACK_ALIGN * YK0A_STACK_ALIGN);
	yk0a_set_priority(3, 1, NULL);
	tool_task_start(3, task3, task3_stack + TASK3_STACK_SIZE / YK0A_STACK_ALIGN * YK0A_STACK_ALIGN);
	yk0a_listen(YK0A_SETUP_EXCEPTION_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
	tool_cons_put_string("{");
	tool_cons_put_hexadecimal(f);
	if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_EXCEPTION_PORT, &t) == 0) {
		yk0a_uint32_t cpsr;
		yk0a_get_context(t, YK0A_CONTEXT_CPSR, &cpsr);
		tool_cons_put_string(" ");
		tool_cons_put_hexadecimal(t);
		tool_cons_put_string(" ");
		tool_cons_put_hexadecimal(cpsr);
	}
	tool_cons_put_string("}\n");
	device_cons_halt();
}

void task1(void) {
	wrap_port_t p;
	for (p = 0; p < WRAP_PORT_SIZE; ++p) {
		tool_queue_initialize((tool_queue_t *)&wrap_port_control[p]);
		wrap_port_control[p].flag = 0;
	}
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		yk0a_listen(YK0A_SETUP_INTERRUPT_PORT, ~YK0A_FLAG_KERNEL_RECEIVER, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && yk0a_accept(YK0A_SETUP_INTERRUPT_PORT, &t) == 0) {
			yk0a_uint32_t r2;
			yk0a_uint32_t r3;
			yk0a_get_context(t, CONTEXT_R2, &r2);
			switch (r2) {
			case WRAP0_ACCEPT:
				yk0a_get_context(t, CONTEXT_R3, &r3);
				if (r3 >= WRAP_PORT_SIZE || (wrap_port_control[r3].flag & YK0A_FLAG_KERNEL_SENDER) == 0) {
					yk0a_set_context(t, CONTEXT_R0, 1, NULL);
					yk0a_reply(t);
				} else {
					yk0a_task_t replyee;
					replyee = (wrap_task_control_t *)tool_queue_remove((tool_queue_t *)&wrap_port_control[r3]) - &wrap_task_control[0];
					if (tool_queue_copy((tool_queue_t *)&wrap_port_control[r3]) == NULL) {
						wrap_port_control[r3].flag &= ~YK0A_FLAG_KERNEL_SENDER;
					}
					yk0a_connect(replyee, t);
					yk0a_set_context(t, CONTEXT_R1, replyee, NULL);
					yk0a_reply(t);
				}
				break;
			case WRAP0_CALL:
				yk0a_get_context(t, CONTEXT_R3, &r3);
				if (r3 >= WRAP_PORT_SIZE) {
					yk0a_set_context(t, CONTEXT_R0, 1, NULL);
					yk0a_reply(t);
				} else {
					tool_queue_append((tool_queue_t *)&wrap_port_control[r3], (tool_queue_t *)&wrap_task_control[t]);
					wrap_port_control[r3].flag |= YK0A_FLAG_KERNEL_SENDER;
					if ((wrap_port_control[r3].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
						yk0a_uint32_t r4;
						yk0a_get_context(wrap_port_control[r3].receiver, CONTEXT_R4, &r4);
						if ((r4 & YK0A_FLAG_KERNEL_SENDER) != 0) {
							yk0a_set_context(wrap_port_control[r3].receiver, CONTEXT_R1, wrap_port_control[r3].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK), NULL);
							yk0a_reply(wrap_port_control[r3].receiver);
							wrap_port_control[r3].flag &= ~YK0A_FLAG_KERNEL_RECEIVER;
						}
					}
				}
				break;
			case WRAP0_LISTEN:
				yk0a_get_context(t, CONTEXT_R3, &r3);
				if (r3 >= WRAP_PORT_SIZE || (wrap_port_control[r3].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
					yk0a_set_context(t, CONTEXT_R0, 1, NULL);
					yk0a_reply(t);
				} else {
					yk0a_uint32_t r4;
					yk0a_get_context(t, CONTEXT_R4, &r4);
					if (((wrap_port_control[r3].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK) | YK0A_FLAG_KERNEL_RECEIVER) & r4) != 0) {
						yk0a_set_context(t, CONTEXT_R1, wrap_port_control[r3].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK) | YK0A_FLAG_KERNEL_RECEIVER, NULL);
						yk0a_reply(t);
						wrap_port_control[0].flag &= ~r4 | YK0A_FLAG_KERNEL_MASK;
					} else {
						wrap_port_control[r3].flag |= YK0A_FLAG_KERNEL_RECEIVER;
						wrap_port_control[r3].receiver = t;
					}
				}
				break;
			default:
				yk0a_set_context(t, CONTEXT_R0, 1, NULL);
				yk0a_reply(t);
				break;
			}
		}
		if ((f & ~YK0A_FLAG_KERNEL_MASK) != 0) {
			wrap_port_control[0].flag |= f & ~YK0A_FLAG_KERNEL_MASK;
			for (p = 0; p < WRAP_PORT_SIZE; ++p) {
				if ((wrap_port_control[p].flag & YK0A_FLAG_KERNEL_RECEIVER) != 0) {
					yk0a_uint32_t r4;
					yk0a_get_context(wrap_port_control[p].receiver, CONTEXT_R4, &r4);
					if ((wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK & r4) != 0) {
						yk0a_set_context(wrap_port_control[p].receiver, CONTEXT_R1, wrap_port_control[p].flag | (wrap_port_control[0].flag & ~YK0A_FLAG_KERNEL_MASK), NULL);
						yk0a_reply(wrap_port_control[p].receiver);
						wrap_port_control[0].flag &= ~r4 | YK0A_FLAG_KERNEL_MASK;
						wrap_port_control[p].flag &= ~YK0A_FLAG_KERNEL_RECEIVER;
					}
				}
			}
		}
	}
}

void task2(void) {
	yk0a_uint8_t c;
	tool_queue_t q;
	yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
	c = 0;
	tool_queue_initialize(&q);
	while (1) {
		yk0a_flag_t f;
		yk0a_task_t t;
		wrap0_listen(YK0A_SETUP_INTERRUPT_PORT, 0, YK0A_FLAG_KERNEL_SENDER | DEVICE_INTERRUPT_CONS, &f);
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0 && wrap0_accept(YK0A_SETUP_INTERRUPT_PORT, 0, &t) == 0) {
			yk0a_uint32_t r4;
			yk0a_uint32_t r5;
			yk0a_get_context(t, CONTEXT_R4, &r4);
			switch (r4) {
			case WRAP1_CONS_GETCHAR:
				tool_queue_append(&q, (tool_queue_t *)&wrap_task_control[t]);
				break;
			case WRAP1_CONS_PUTCHAR:
				yk0a_get_context(t, CONTEXT_R5, &r5);
				yk0a_reply(t);
				device_cons_putchar(r5);
				break;
			default:
				yk0a_forward(YK0A_SETUP_EXCEPTION_PORT, t);
				break;
			}
		}
		if ((f & DEVICE_INTERRUPT_CONS) != 0) {
			c = device_cons_getchar();
			yk0a_interrupt_bit_mask(0, DEVICE_INTERRUPT_CONS, NULL);
		}
		if (c != 0 && tool_queue_copy(&q) != NULL) {
			t = (wrap_task_control_t *)tool_queue_remove(&q) - &wrap_task_control[0];
			yk0a_set_context(t, CONTEXT_R1, c, NULL);
			yk0a_reply(t);
			c = 0;
		}
	}
}

void task3(void) {
	while (1) {
		yk0a_uint8_t c;
		wrap1_cons_getchar(YK0A_SETUP_INTERRUPT_PORT, 0, &c);
		if (c < ' ') {
			break;
		}
		wrap1_cons_putchar(YK0A_SETUP_INTERRUPT_PORT, 0, c);
	}
	wrap1_cons_halt(YK0A_SETUP_INTERRUPT_PORT, 0);
}
